from datetime import date, datetime
import pytest
from pyspark.sql import DataFrame, Row, SparkSession
from delta import configure_spark_with_delta_pip


@pytest.fixture(scope="session")
def spark():
    builder = (
        SparkSession.builder.appName("MyApp")
        .config("spark.sql.extensions", "io.delta.sql.DeltaSparkSessionExtension")
        .config(
            "spark.sql.catalog.spark_catalog", "org.apache.spark.sql.delta.catalog.DeltaCatalog"
        )
    )

    spark = configure_spark_with_delta_pip(builder).getOrCreate()

    return spark


@pytest.fixture(scope="session")
def nested_dataset(spark: SparkSession) -> DataFrame:
    return spark.sql(
        """
        select
            1 as column_a
        ,   2.0 as column_b
        ,   struct(
                "c1" as column_c_1
            ,   struct(
                    "c_2_1" as c_2_1,
                    "c_2_2" as c_2_2,
                    "c_2_3" as c_2_3
            ) as column_c_2
        ) as column_c
        """
    )


@pytest.fixture(scope="session")
def simple_delta_dataset(spark: SparkSession) -> DataFrame:

    return spark.createDataFrame(
        [
            Row(a=1, b=2.0, c="string1", d=date(2000, 1, 1), e=datetime(2000, 1, 1, 12, 0)),
            Row(a=2, b=3.0, c="string2", d=date(2000, 2, 1), e=datetime(2000, 1, 2, 12, 0)),
        ]
    )


@pytest.fixture(scope="session")
def flat_dataset(spark: SparkSession) -> DataFrame:
    return spark.createDataFrame(
        [
            Row(
                a=1,
                b=2.0,
                c="string1",
                d=date(2000, 1, 1),
                e=datetime(2000, 1, 1, 12, 0),
            ),
            Row(
                a=2,
                b=3.0,
                c="string2",
                d=date(2000, 2, 1),
                e=datetime(2000, 1, 2, 12, 0),
            ),
        ]
    )


@pytest.fixture(scope="session")
def nested_flat_array(spark: SparkSession) -> DataFrame:
    return spark.createDataFrame(
        Row(
            value=Row(
                RetailTransaction=Row(
                    StoreID="1234",
                    LineItem=["Card", "Discount"],
                )
            )
        )
    )


@pytest.fixture(scope="session")
def nested_array_struct(spark: SparkSession) -> DataFrame:
    return spark.createDataFrame(
        Row(
            value=Row(
                RetailTransaction=Row(
                    StoreID="1234",
                    LineItem=[
                        Row(
                            SequenceNumber="2",
                            LoyaltyCard=Row(CardId="111", CardType="Premium"),
                        ),
                        Row(
                            SequenceNumber="3",
                            LoyaltyCard=Row(CardId="222", CardType="Gold"),
                        ),
                    ],
                )
            )
        )
    )


@pytest.fixture(scope="session")
def nested_array_array(spark: SparkSession) -> DataFrame:
    return spark.createDataFrame(
        Row(
            value=Row(
                RetailTransaction=Row(
                    StoreID="1234",
                    LineItem=[
                        Row(
                            Discount=[
                                Row(POSItemId="100", POSName="POS01"),
                                Row(POSItemId="101", POSName="POS02"),
                            ]
                        )
                    ],
                )
            )
        )
    )


@pytest.fixture(scope="session")
def nested_array_struct_array(spark: SparkSession) -> DataFrame:
    return spark.createDataFrame(
        Row(
            value=Row(
                RetailTransaction=Row(
                    StoreID="1234",
                    LineItem=[
                        Row(
                            Discount=Row(
                                DiscountId="1",
                                ItemList=[Row(ItemID="10", TestId="200")],
                            )
                        )
                    ],
                )
            )
        )
    )


@pytest.fixture(scope="session")
def nested_array_struct_array_multiple_rows(spark: SparkSession) -> DataFrame:
    return spark.createDataFrame(
        Row(
            value=Row(
                RetailTransaction=Row(
                    StoreID="1234",
                    LineItem=[
                        Row(
                            Discount=Row(
                                DiscountId="1",
                                ItemList=[Row(ItemID="10", ItemNum="11")],
                            )
                        ),
                        Row(
                            Discount=Row(
                                DiscountId="2",
                                ItemList=[Row(ItemID="12", ItemNum="13")],
                            )
                        ),
                    ],
                )
            )
        )
    )


@pytest.fixture(scope="session")
def nested_array_struct_struct(spark: SparkSession) -> DataFrame:
    return spark.createDataFrame(
        Row(
            value=Row(
                RetailTransaction=Row(
                    StoreID="1234",
                    LineItem=[
                        Row(
                            LoyaltyItem=Row(
                                SequenceNumber="2",
                                LoyaltyCard=Row(CardId="2222", CardType="Premium"),
                            )
                        )
                    ],
                )
            )
        )
    )


@pytest.fixture(scope="session")
def nested_tlog_example(spark: SparkSession) -> DataFrame:
    return spark.createDataFrame(
        Row(
            value=Row(
                RetailTransaction=Row(
                    StoreID="1234",
                    LineItem=[
                        Row(
                            LoyaltyItem=Row(
                                SequenceNumber="2",
                                LoyaltyCard=Row(CardId="2222", CardType="Premium"),
                            ),
                            Discount=[Row(POSItemId="100", POSName="POS01")],
                        )
                    ],
                )
            )
        )
    )


@pytest.fixture(scope="session")
def nested_root_array(spark: SparkSession) -> DataFrame:
    return spark.createDataFrame(
        Row(value=Row(RetailTransaction=[Row(StoreID="1234", StoreName="Test Store")]))
    )



from pyspark.sql import DataFrame
from pyspark.sql import functions as F
import pytest
from spalah.dataframe import SchemaComparer, MatchedColumn, NotMatchedColumn


def test_SchemaComparer_equal_schema(flat_dataset: DataFrame) -> None:
    """Two equal schemas must result into all matched columns"""

    df1 = flat_dataset

    schema_comparer = SchemaComparer(source_schema=df1.schema, target_schema=df1.schema)

    schema_comparer.compare()

    assert len(schema_comparer.matched) > 0
    assert len(schema_comparer.not_matched) == 0


def test_SchemaComparer_compare_different_schemas(flat_dataset: DataFrame) -> None:
    """Two equal schemas must result into all matched columns"""

    df1 = flat_dataset

    df2 = (
        flat_dataset.withColumn("z", F.lit("z"))  # add extra column z
        .drop("a", "b")  # drop columns a, b
        .withColumn("b", F.lit("b"))  # recreate column b with a string type
        .withColumnRenamed("c", "C")  # change the case of column: c=>C
    )

    schema_comparer = SchemaComparer(source_schema=df1.schema, target_schema=df2.schema)

    schema_comparer.compare()

    assert set(schema_comparer.matched) == {
        MatchedColumn(name="d", data_type="DateType()"),
        MatchedColumn(name="e", data_type="TimestampType()"),
    }

    assert set(schema_comparer.not_matched) == {
        NotMatchedColumn(
            name="c",
            data_type="StringType()",
            reason="The column exists in source and target schemas but it's name "
            "is case-mismatched",
        ),
        NotMatchedColumn(
            name="b",
            data_type="DoubleType() <=> StringType()",
            reason="The column exists in source and target schemas but it is not "
            "matched by a data type",
        ),
        NotMatchedColumn(
            name="a", data_type="LongType()", reason="The column exists only in the source schema"
        ),
        NotMatchedColumn(
            name="z",
            data_type="StringType()",
            reason="The column exists only in the target schema",
        ),
    }


def test_SchemaComparer_compare_exception_passed_dataframe(flat_dataset: DataFrame) -> None:
    """Two equal schemas must result into all matched columns"""

    df1 = flat_dataset

    with pytest.raises(Exception) as e:
        schema_comparer = SchemaComparer(
            source_schema=df1.schema,
            target_schema=df1,  # Dataframe passed instead of StructType (schema)
        )
        schema_comparer.compare()
    assert str(e.value).startswith(
        "One of 'source_schema or 'target_schema' passed as a DataFrame."
    )


def test_SchemaComparer_compare_exception_passed_string(flat_dataset: DataFrame) -> None:
    """Two equal schemas must result into all matched columns"""

    df1 = flat_dataset

    with pytest.raises(Exception) as e:
        schema_comparer = SchemaComparer(
            source_schema=df1.schema,
            target_schema="abcde",  # String passed instead of StructType (schema)
        )
        schema_comparer.compare()
    assert str(e.value).startswith(
        "Parameters 'source_schema and 'target_schema' " "must have a type: StructType"
    )



import copy
from pprint import pformat, pprint
from typing import Union, List, Set

from pyspark.sql import DataFrame
from pyspark.sql import functions as F
from pyspark.sql import types as T
from collections import namedtuple


MatchedColumn = namedtuple("MatchedColumn", ["name", "data_type"])
NotMatchedColumn = namedtuple("NotMatchedColumn", ["name", "data_type", "reason"])


def __remove_suffix(input_string: str, suffix: str) -> str:
    """Remove suffix from string if it exists

    Args:
        input_string (str):  Input string
        suffix (str):  Suffix to remove

    Returns:
        str: Input string without suffix
    """
    if suffix and input_string.endswith(suffix):
        return input_string[: -len(suffix)]
    return input_string


def __process_schema_node(
    node: T.StructField,
    column_prefix: str = "",
    columns_to_include: Union[List, None] = None,  # type: ignore
    columns_to_exclude: Union[List, None] = None,  # type: ignore
    nullify_only: bool = False,
    array_element: Union[None, F.col] = None,
    debug: bool = False,
) -> Union[F.col, F.lit, None]:
    """Internal function to process a node of the schema

    Args:
        node (dict): json representation of a certain node in schema
        column_prefix (str, optional): starting path of the nested column name
        columns_to_include (list): Columns that must remain in the dataframe unchanged
        columns_to_exclude (list): Columns that must be removed (or nullified)
        nullify_only (bool, optional): If True, only nullify columns, do not remove them
        array_element (Union[None, F.col], optional): reference to the array element if
            it must be transformed
        debug (bool, optional): If True, print debug information, defaults to False

    Returns:
        Union[Column, None]: The pyspark column instance of a certain node of schema
        if column to be rendered or None if column to be removed from the final projection
    """

    # Default values
    col_expression = F.lit(-1)
    node_name = node.name
    is_struct_without_children = False
    is_struct = True

    if column_prefix == "":
        column_prefix = node_name
    else:
        column_prefix = column_prefix + "." + node_name
    column_prefix_dot = f"{column_prefix}.".lower()
    include_this_node = False

    # Process exclude/include rules
    if columns_to_exclude and any(
        [
            x
            for x in [f"{x}." for x in columns_to_exclude]
            if x == column_prefix_dot or column_prefix_dot.startswith(x)
        ]
    ):
        include_this_node = False
    # If no columns to include specified, then filter out other columns
    elif columns_to_include and any(
        [
            x
            for x in [f"{x}." for x in columns_to_include]
            if x.startswith(column_prefix_dot) or column_prefix_dot.startswith(x)
        ]
    ):
        include_this_node = True
    # If no columns to include specified, then all columns are accepted
    elif not columns_to_include or len(columns_to_include) == 0:
        include_this_node = True

    # Structs
    if isinstance(node.dataType, T.StructType):

        children = []

        for field in node.dataType.fields:

            if isinstance(field.dataType, T.StructType) and array_element is not None:
                array_col_child = array_element[field.name]
            elif array_element is not None:
                array_col_child = array_element
            else:
                array_col_child = None

            _child = __process_schema_node(
                node=field,
                column_prefix=column_prefix,
                columns_to_include=columns_to_include,
                columns_to_exclude=columns_to_exclude,
                nullify_only=nullify_only,
                array_element=array_col_child,
                debug=debug,
            )
            if _child is not None:
                children.append(_child)

        if len(children) == 0:
            is_struct_without_children = True
        col_expression = F.struct(children).alias(node_name)

    # Struct within array (or: array(struct()))
    elif (
        isinstance(node.dataType, T.ArrayType)
        and isinstance(node.dataType.elementType, T.StructType)
        and
        # to include this node only when it must appear in the final projection
        (nullify_only or include_this_node)
    ):

        children = []

        def _transform_array(array_element: F.col) -> F.col:
            """Internal function to process elements of an array
            one by one and return the transformed array

            Args:
                row (F.col): array element to transform

            Returns:
                F.col: transformed array element
            """

            # converts schema node array(struct()) -> struct()
            struct_in_array_node = copy.deepcopy(node)
            struct_in_array_node.dataType = node.dataType.elementType

            # removes the name of the array node from the element path
            # because the existence of the array does not create a new level in the path
            column_prefix_new = __remove_suffix(__remove_suffix(column_prefix, node_name), ".")

            struct_extracted_from_array_element = __process_schema_node(
                node=struct_in_array_node,
                column_prefix=column_prefix_new,
                columns_to_include=columns_to_include,
                columns_to_exclude=columns_to_exclude,
                nullify_only=nullify_only,
                array_element=array_element,
                debug=debug,
            )
            return struct_extracted_from_array_element

        # if this is first iteration (or level of arrays) then
        # the array path is the element path
        # otherwise, the array path is child node of the parent array
        if array_element is None:
            array_path = column_prefix
        else:
            array_path = array_element[node_name]

        col_expression = F.transform(array_path, lambda x: _transform_array(array_element=x)).alias(
            node_name
        )

    # Regular columns
    else:

        # get node data type
        # In case of nullification of array it converted to a string null value
        if isinstance(node.dataType, T.ArrayType):
            node_type = T.StringType()  # arrays will be nullified and casted as strings
        else:
            node_type = node.dataType

        # regular column expression, valid for non-nullified columns
        if array_element is None:
            col_expression = F.col(column_prefix).alias(node_name)
        else:
            col_expression = array_element[node_name].alias(node_name)

        # column expression for nullified fields
        nullified_col_expression = F.lit(None).cast(node_type).alias(node_name)

        # flag for further processing that the field is not struct
        is_struct = False

    if include_this_node and debug:
        print(f" - {column_prefix}")

    if include_this_node and not is_struct_without_children:
        return col_expression
    if not include_this_node and nullify_only:

        if is_struct:
            return col_expression
        else:
            return nullified_col_expression
    else:
        return None


def slice_dataframe(
    input_dataframe: DataFrame,
    columns_to_include: Union[List, None] = None,
    columns_to_exclude: Union[List, None] = None,
    nullify_only: bool = False,
    debug: bool = False,
) -> DataFrame:
    """Process flat or nested schema of the dataframe by slicing the schema
    or nullifying columns

    Args:
        input_dataframe (DataFrame): Input dataframe
        columns_to_include (list): Columns that must remain in the dataframe unchanged
        columns_to_exclude (list): Columns that must be removed (or nullified)
        nullify_only (bool, optional): Nullify columns instead of removing them. Defaults to False
        debug (bool, optional): For extra debug output. Defaults to False.

    Raises:
        TypeError: If the 'column_to_include' or 'column_to_exclude' are not lists
        ValueError: If the included columns overlay excluded columns, so nothing to return

    Returns:
        DataFrame: The processed dataframe
    """

    projection = []
    _schema = input_dataframe.schema

    if not columns_to_include:
        columns_to_include = []

    if not columns_to_exclude:
        columns_to_exclude = []

    if not (type(columns_to_include) is list and type(columns_to_exclude) is list):
        raise TypeError("'columns_to_include' and 'columns_to_exclude' must be a list")

    if not all(isinstance(item, str) for item in columns_to_include + columns_to_exclude):
        raise TypeError("Member of 'columns_to_include' and 'columns_to_exclude' must be a string")

    if debug:
        print("The list of columns to include:")
        pprint(columns_to_include)

        print("The list of columns to exclude:")
        pprint(columns_to_exclude)

        if nullify_only:
            print("Columns to nullify in the final projection:")
        else:
            print("Columns to include into the final projection:")

    # lower-case items for making further filtering case-insensitive
    columns_to_include = [item.lower() for item in columns_to_include]
    columns_to_exclude = [item.lower() for item in columns_to_exclude]

    for field in _schema.fields:
        node_result = __process_schema_node(
            node=field,
            columns_to_include=columns_to_include,
            columns_to_exclude=columns_to_exclude,
            nullify_only=nullify_only,
            debug=debug,
        )

        if node_result is not None:
            projection.append(node_result)

    result = input_dataframe.select(*projection)

    if len(result.columns) == 0:
        raise ValueError(
            "At least one column should be listed in the "
            + "columns_to_include/columns_to_exclude attributes "
            + "and included column should not directly overlap with excluded one"
        )

    return result


def flatten_schema(
    schema: T.StructType,
    include_datatype: bool = False,
    column_prefix: Union[str, None] = None,
) -> list:
    """Generates the flatten list of columns from the complex dataframe schema

    Args:
        schema (StructType): input dataframe's schema
        include_type (bool, optional): Include column types
        column_prefix (str, optional): The column name prefix. Defaults to None.

    Returns:
        The list of (flattened) column names
    """

    if not isinstance(schema, T.StructType):
        raise TypeError("Parameter schema must be a StructType")

    columns = []

    for column in schema.fields:
        if column_prefix:
            name = column_prefix + "." + column.name
        else:
            name = column.name

        column_data_type = column.dataType

        if isinstance(column_data_type, T.ArrayType):
            column_data_type = column_data_type.elementType

        if isinstance(column_data_type, T.StructType):
            columns += flatten_schema(
                column_data_type,
                include_datatype=include_datatype,
                column_prefix=name,
            )
        else:
            if include_datatype:
                result = name, str(column_data_type)
            else:
                result = name

            columns.append(result)

    return columns


def script_dataframe(input_dataframe: DataFrame, suppress_print_output: bool = True) -> str:

    """Generates a script of a dataframe"""

    MAX_ROWS_IN_SCRIPT = 20

    __dataframe = input_dataframe

    if __dataframe.count() > MAX_ROWS_IN_SCRIPT:
        raise ValueError(
            "This method is limited to script up " f"to {MAX_ROWS_IN_SCRIPT} row(s) per call"
        )

    __schema = input_dataframe.schema.jsonValue()

    __script_lines = [
        "from pyspark.sql import Row",
        "import datetime",
        "from decimal import Decimal",
        "from pyspark.sql.types import *",
        "",
        "# Scripted data and schema:",
        f"__data = {pformat(__dataframe.collect())}",
        "",
        f"__schema = {__schema}",
        "",
        "outcome_dataframe = spark.createDataFrame(__data, StructType.fromJson(__schema))",
    ]

    __final_script = "\n".join(__script_lines)

    if not suppress_print_output:
        print("#", "=" * 80)
        print(
            "# IMPORTANT!!! REMOVE PII DATA BEFORE RE-CREATING IT IN NON-PRODUCTION ENVIRONMENTS",
            " " * 3,
            "#",
        )
        print("#", "=" * 80)
        print("")
        print(__final_script)

    return __final_script


class SchemaComparer:
    def __init__(self, source_schema: T.StringType, target_schema: T.StringType) -> None:
        self._source = self.__import_schema(source_schema)
        self._target = self.__import_schema(target_schema)
        self.matched: List[tuple] = list()
        self.not_matched: List[tuple] = list()

    def __import_schema(self, input_schema: T.StructType) -> Set[tuple]:
        """Import StructType as the flatten set of tuples: (column_name, data_type)

        Args:
            input_schema (T.StructType): Schema to process

        Raises:
            TypeError: if input schema has a type: DataFrame
            TypeError: if input schema hasn't a type: StructType

        Returns:
            Set[tuple]: Set of tuples: (column_name, data_type)
        """

        if isinstance(input_schema, DataFrame):
            raise TypeError(
                "One of 'source_schema or 'target_schema' passed as a DataFrame. "
                "Use DataFrame.schema instead"
            )
        elif not isinstance(input_schema, T.StructType):
            raise TypeError(
                "Parameters 'source_schema and 'target_schema' must have a type: StructType"
            )

        return set(flatten_schema(input_schema, True))

    def __match_by_name_and_type(
        self, source: Set[tuple] = set(), target: Set[tuple] = set()
    ) -> Set[tuple]:
        """Matches columns in source and target schemas by name and data type

        Args:
            source (Set[tuple], optional): Flattened source schema. Defaults to set().
            target (Set[tuple], optional): Flattened target schema. Defaults to set().

        Returns:
            Set[tuple]: Fully matched columns as a set of tuples: (column_name, data_type)
        """

        # If source and target is not provided, use class attributes as the input
        _source = self._source if not source else source
        _target = self._target if not target else target

        result = _source & _target

        # Remove matched values of case 1 from further processing
        self.__remove_matched_by_name_and_type(result)

        if not (source and target):
            self.__populate_matched(result)

        return result

    def __remove_matched_by_name_and_type(self, subtract_value: Set[tuple]) -> None:
        """Removes fully matched columns from the further processing

        Args:
            subtract_value (Set[tuple]): Set of matched columns
        """

        self._source = self._source - subtract_value
        self._target = self._target - subtract_value

    def __remove_matched_by_name(self, subtract_value: Set[tuple]) -> None:
        """Removes matched by name columns from the further processing

        Args:
            subtract_value (Set[tuple]): Set of matched column
        """

        def _remove(input_value: Set[tuple], subtract_value: Set[tuple]) -> Set[tuple]:
            """Internal helper for removal of Tuples by the first member"""
            return {
                (x, y)
                for (x, y) in input_value
                if not x.lower() in ([z[0].lower() for z in subtract_value])
            }

        self._source = _remove(self._source, subtract_value)  # type: ignore
        self._target = _remove(self._target, subtract_value)  # type: ignore

    def __lower_column_names(self, base_value: Set[tuple]) -> Set[tuple]:
        """Lower-case all column names of the input set

        Args:
            base_value (Set[tuple]): Input set of columns

        Returns:
            Set[tuple]: Output set of columns with lower-case column names
        """
        return {(x.lower(), y) for (x, y) in base_value}

    def __match_by_name_type_excluding_case(self) -> None:
        """Matches columns in source and target schemas by name and data type
        without taking into account column name case
        """

        _source_lowered = self.__lower_column_names(self._source)
        _target_lowered = self.__lower_column_names(self._target)

        result = self.__match_by_name_and_type(_source_lowered, _target_lowered)

        # Remove matched values of case 2 from further processing
        self.__remove_matched_by_name(result)

        self.__populate_not_matched(
            result,
            "The column exists in source and target schemas but it's name is case-mismatched",
        )

    def __match_by_name_but_not_type(self) -> None:
        """Matches columns in source and target schemas only by column name"""

        x = dict(self._source)  # type: ignore
        y = dict(self._target)  # type: ignore

        result = {(k, f"{x[k]} <=> {y[k]}") for k in x if k in y and x[k] != y[k]}

        # Remove matched values of case 3 from further processing
        self.__remove_matched_by_name(result)  # type: ignore

        self.__populate_not_matched(
            result,  # type: ignore
            "The column exists in source and target schemas but it is not matched by a data type",
        )

    def __process_remaining_non_matched_columns(self) -> None:
        """Process remaining not matched columns"""

        self.__populate_not_matched(self._source, "The column exists only in the source schema")

        self.__populate_not_matched(self._target, "The column exists only in the target schema")

        self.__remove_matched_by_name(self._source)
        self.__remove_matched_by_name(self._target)

    def __populate_matched(self, input_value: Set[tuple]) -> None:
        """Populate class property 'matched' with a list of fully matched columns

        Args:
            input_value (Set[tuple]): The set of tuples with a list of column names and data types
        """

        for match in input_value:
            self.matched.append(MatchedColumn(name=match[0], data_type=match[1]))

    def __populate_not_matched(self, input_value: Set[tuple], reason: str) -> None:
        """Populate class property 'not_matched' with a list of columns that didn't match for some
        reason with included an actual reason

        Args:
            input_value (Set[tuple]): The set of tuples with a list of column names and data types
            reason (str): Reason for not match
        """

        for match in input_value:
            self.not_matched.append(
                NotMatchedColumn(name=match[0], data_type=match[1], reason=reason)
            )

    def compare(self) -> None:
        """
        Compares the source and target schemas and populates properties
        'matched' and 'not_matched'
        """

        # Case 1: find columns that are matched by name and type and remove them
        # from further processing
        self.__match_by_name_and_type()

        # Case 2: find columns that match mismatched by name due to case: ID <-> Id
        self.__match_by_name_type_excluding_case()

        # Case 3: Find columns matched by name, but not by data type
        self.__match_by_name_but_not_type()

        # Case 4: Find columns that exists only in the source or target
        self.__process_remaining_non_matched_columns()



import pytest
from spalah.datalake import set_delta_properties, get_delta_properties


def test_set_table_properties_new_table_no_existing_properties(spark, tmp_path) -> None:
    """Tested method must set correctly table properties to the brand new dataset"""

    delta_path = f"{tmp_path}/sample_delta_dataset"

    properties_to_set = {
        "delta.logRetentionDuration": "interval 10 days",
        "delta.deletedFileRetentionDuration": "interval 16 days",
    }

    spark.range(0, 1).write.format("delta").mode("overwrite").save(delta_path)

    set_delta_properties(table_path=delta_path, properties=properties_to_set)

    get_properties = get_delta_properties(table_path=delta_path)

    assert properties_to_set == get_properties


def test_set_table_properties_new_table_with_existing_properties(spark, tmp_path) -> None:
    """Tested method must set correctly table properties to the brand new dataset.
    An existing property must be preserved
    """

    delta_path = f"{tmp_path}/sample_delta_dataset"

    properties_to_set = {
        "delta.logRetentionDuration": "interval 10 days",
        "delta.deletedFileRetentionDuration": "interval 16 days",
    }

    spark.range(0, 1).write.format("delta").mode("overwrite").save(delta_path)

    _sql = f"ALTER TABLE delta.`{delta_path}` SET TBLPROPERTIES (prop.property1 = 'abc')"
    spark.sql(_sql)

    set_delta_properties(table_path=delta_path, properties=properties_to_set)

    get_properties = get_delta_properties(table_path=delta_path)

    properties_to_set.update({"prop.property1": "abc"})

    assert get_properties == properties_to_set


def test_set_table_properties_new_table_remove_existing(spark, tmp_path) -> None:
    """Tested method must set correctly table properties to the brand new dataset.
    An existing property must be removed because of option allow_unset=True,
    therefore only the table options defined in the input dictionary must be set"""

    delta_path = f"{tmp_path}/sample_delta_dataset"

    properties_to_set = {
        "delta.logRetentionDuration": "interval 10 days",
        "delta.deletedFileRetentionDuration": "interval 16 days",
    }

    spark.range(0, 1).write.format("delta").mode("overwrite").save(delta_path)

    _sql = f"ALTER TABLE delta.`{delta_path}` SET TBLPROPERTIES (prop.property1 = 'abc')"
    spark.sql(_sql)

    set_delta_properties(table_path=delta_path, properties=properties_to_set, keep_existing=False)

    get_properties = get_delta_properties(table_path=delta_path)

    assert get_properties == properties_to_set


def test_set_table_properties_exceptions_both_provided_as_identifier() -> None:
    "When both table_path and table_name provided the exception must occur"

    with pytest.raises(ValueError) as e:
        set_delta_properties(
            table_path="/some/path",
            table_name="tbl_name",
            properties={"a": "b"},
        )
    assert str(e.value).startswith("Both 'table_path' and 'table_name' provided")


def test_set_table_properties_exceptions_noting_provided_as_identifier() -> None:
    "When neither table_path nor table_name provided the exception must occur"

    with pytest.raises(ValueError) as e:
        set_delta_properties(
            table_path="",
            table_name="",
            properties={"a": "b"},
        )
    assert str(e.value).startswith("Neither 'table_path' nor 'table_name' provided")



import pytest
from pyspark.sql import Row

from spalah.dataframe import slice_dataframe


@pytest.mark.parametrize(
    "assert_message,input_dataset,columns_to_include,columns_to_exclude,nullify_only,expected",
    [
        # nested structs: include
        (
            "The dataframe must contain only columns listed in "
            "'columns_to_include' (nested structs)",
            "nested_dataset",
            ["column_a", "column_c"],
            [],
            False,
            Row(
                column_a=1,
                column_c=Row(
                    column_c_1="c1", column_c_2=Row(c_2_1="c_2_1", c_2_2="c_2_2", c_2_3="c_2_3")
                ),
            ),
        ),
        # nested structs: include, mixed case
        (
            "The dataframe must contain only columns listed in "
            "'columns_to_include' (nested structs)",
            "nested_dataset",
            ["Column_A", "column_c"],
            [],
            False,
            Row(
                column_a=1,
                column_c=Row(
                    column_c_1="c1", column_c_2=Row(c_2_1="c_2_1", c_2_2="c_2_2", c_2_3="c_2_3")
                ),
            ),
        ),
        # "nested structs: include and exclude"
        (
            "The dataframe must contain columns listed in "
            "'columns_to_include' beside of 'columns_to_exclude'  (nested structs)",
            "nested_dataset",
            ["column_b", "column_c"],
            ["column_c.column_c_2.c_2_1"],
            False,
            Row(
                column_b=2.0,
                column_c=Row(column_c_1="c1", column_c_2=Row(c_2_2="c_2_2", c_2_3="c_2_3")),
            ),
        ),
        # "nested structs: exclude"
        (
            "All columns in 'columns_to_exclude' must be excluded (nested structs)'",
            "nested_dataset",
            [],
            ["column_b", "column_c.column_c_2.c_2_1"],
            False,
            Row(
                column_a=1,
                column_c=Row(column_c_1="c1", column_c_2=Row(c_2_2="c_2_2", c_2_3="c_2_3")),
            ),
        ),
        # "nested structs: exclude, mixed case"
        (
            "All columns in 'columns_to_exclude' must be excluded (nested structs)'",
            "nested_dataset",
            [],
            ["Column_B", "Column_C.column_c_2.C_2_1"],
            False,
            Row(
                column_a=1,
                column_c=Row(column_c_1="c1", column_c_2=Row(c_2_2="c_2_2", c_2_3="c_2_3")),
            ),
        ),
        # "nested structs: nullify only for excluded columns"
        (
            "All columns in 'columns_to_exclude' must be nullified (nested structs)'",
            "nested_dataset",
            [],
            ["column_b", "column_c.column_c_2.c_2_1"],
            True,
            Row(
                column_a=1,
                column_b=None,
                column_c=Row(
                    column_c_1="c1", column_c_2=Row(c_2_1=None, c_2_2="c_2_2", c_2_3="c_2_3")
                ),
            ),
        ),
        # "flat: nullify only for excluded columns"
        (
            "All columns in 'columns_to_exclude' must be nullified (flat)'",
            "flat_dataset",
            [],
            ["d", "e"],
            True,
            Row(a=1, b=2.0, c="string1", d=None, e=None),
        ),
        # "flat: exclude"
        (
            "All columns in 'columns_to_exclude' must be excluded (flat)'",
            "flat_dataset",
            [],
            ["d", "e"],
            False,
            Row(a=1, b=2.0, c="string1"),
        ),
        # "nested: nullify flat ArrayType"
        (
            "All columns in 'columns_to_exclude' must be nullified (nested)'",
            "nested_flat_array",
            [],
            ["RetailTransaction.LineItem"],
            True,
            Row(RetailTransaction=Row(StoreID="1234", LineItem=None)),
        ),
        # "nested: nullify fields within ArrayType-Struct"
        (
            "All columns in 'columns_to_exclude' must be nullified (nested)'",
            "nested_array_struct",
            [],
            ["RetailTransaction.LineItem.LoyaltyCard.CardId"],
            True,
            Row(
                RetailTransaction=Row(
                    StoreID="1234",
                    LineItem=[
                        Row(
                            SequenceNumber="2",
                            LoyaltyCard=Row(CardId=None, CardType="Premium"),
                        ),
                        Row(
                            SequenceNumber="3",
                            LoyaltyCard=Row(CardId=None, CardType="Gold"),
                        ),
                    ],
                )
            ),
        ),
        # "nested: nullify Root of ArrayType-Struct"
        (
            "All columns in 'columns_to_exclude' must be nullified (nested)'",
            "nested_array_struct",
            [],
            ["RetailTransaction.LineItem"],
            True,
            Row(
                RetailTransaction=Row(
                    StoreID="1234",
                    LineItem=[
                        Row(
                            SequenceNumber=None,
                            LoyaltyCard=Row(CardId=None, CardType=None),
                        ),
                        Row(
                            SequenceNumber=None,
                            LoyaltyCard=Row(CardId=None, CardType=None),
                        ),
                    ],
                )
            ),
        ),
        # "nested: nullify fields within ArrayType-Struct-ArrayType"
        (
            "All columns in 'columns_to_exclude' must be nullified (nested)'",
            "nested_array_struct_array",
            [],
            ["RetailTransaction.LineItem.Discount.ItemList.ItemID"],
            True,
            Row(
                RetailTransaction=Row(
                    StoreID="1234",
                    LineItem=[
                        Row(
                            Discount=Row(
                                DiscountId="1",
                                ItemList=[Row(ItemID=None, TestId="200")],
                            )
                        )
                    ],
                )
            ),
        ),
        # "nested: nullify fields within ArrayType-Struct-ArrayType multiple rows"
        (
            "All columns in 'columns_to_exclude' must be nullified (nested)'",
            "nested_array_struct_array_multiple_rows",
            [],
            ["RetailTransaction.LineItem.Discount.ItemList.ItemID"],
            True,
            Row(
                RetailTransaction=Row(
                    StoreID="1234",
                    LineItem=[
                        Row(
                            Discount=Row(
                                DiscountId="1",
                                ItemList=[Row(ItemID=None, ItemNum="11")],
                            )
                        ),
                        Row(
                            Discount=Row(
                                DiscountId="2",
                                ItemList=[Row(ItemID=None, ItemNum="13")],
                            )
                        ),
                    ],
                )
            ),
        ),
        # "nested: nullify fields within ArrayType-ArrayType",
        (
            "All columns in 'columns_to_exclude' must be nullified (nested)'",
            "nested_array_array",
            [],
            ["RetailTransaction.LineItem.Discount.POSItemId"],
            True,
            Row(
                RetailTransaction=Row(
                    StoreID="1234",
                    LineItem=[
                        Row(
                            Discount=[
                                Row(POSItemId=None, POSName="POS01"),
                                Row(POSItemId=None, POSName="POS02"),
                            ]
                        )
                    ],
                )
            ),
        ),
        # "nested: nullify fields within ArrayType-Struct-Struct"
        (
            "All columns in 'columns_to_exclude' must be nullified (nested)'",
            "nested_array_struct_struct",
            [],
            ["RetailTransaction.LineItem.LoyaltyItem.LoyaltyCard.CardId"],
            True,
            Row(
                RetailTransaction=Row(
                    StoreID="1234",
                    LineItem=[
                        Row(
                            LoyaltyItem=Row(
                                SequenceNumber="2",
                                LoyaltyCard=Row(CardId=None, CardType="Premium"),
                            )
                        )
                    ],
                )
            ),
        ),
        # "nested: exclude fields in nested arrays"
        (
            "All columns in 'columns_to_exclude' must be nullified (nested)'",
            "nested_tlog_example",
            [],
            [
                "RetailTransaction.LineItem.LoyaltyItem.LoyaltyCard.CardId",
                "RetailTransaction.LineItem.Discount.POSName",
            ],
            False,
            Row(
                RetailTransaction=Row(
                    StoreID="1234",
                    LineItem=[
                        Row(
                            LoyaltyItem=Row(
                                SequenceNumber="2",
                                LoyaltyCard=Row(CardType="Premium"),
                            ),
                            Discount=[Row(POSItemId="100")],
                        )
                    ],
                )
            ),
        ),
        # "nested: nullify fields within array as root"
        (
            "All columns in 'columns_to_exclude' must be nullified (nested)'",
            "nested_root_array",
            [],
            ["RetailTransaction.StoreID"],
            True,
            Row(RetailTransaction=[Row(StoreID=None, StoreName="Test Store")]),
        ),
    ],
    ids=[
        "nested structs: include",
        "nested structs: include, mixed case",
        "nested structs: include and exclude",
        "nested structs: exclude",
        "nested structs: exclude, mixed case",
        "nested structs: nullify only for excluded columns",
        "flat: nullify only for excluded columns",
        "flat: exclude",
        "nested: nullify flat ArrayType",
        "nested: nullify fields within ArrayType-Struct",
        "nested: nullify Root of ArrayType-Struct",
        "nested: nullify fields within ArrayType-Struct-ArrayType",
        "nested: nullify fields within ArrayType-Struct-ArrayType multiple rows",
        "nested: nullify fields within ArrayType-ArrayType",
        "nested: nullify fields within ArrayType-Struct-Struct",
        "nested: exclude fields in nested arrays",
        "nested: nullify fields within array as root",
    ],
)
def test_slice_dataframe(
    assert_message,
    input_dataset,
    columns_to_include,
    columns_to_exclude,
    nullify_only,
    expected,
    request,
) -> None:

    dataset = request.getfixturevalue(input_dataset)

    actual = slice_dataframe(
        input_dataframe=dataset,
        columns_to_include=columns_to_include,
        columns_to_exclude=columns_to_exclude,
        nullify_only=nullify_only,
        debug=True,
    )

    assert actual.first() == expected, assert_message


def test_slice_dataframe_invalid_parameters(request):
    dataset = request.getfixturevalue("flat_dataset")

    with pytest.raises(Exception) as e:
        slice_dataframe(
            input_dataframe=dataset, columns_to_include="a", columns_to_exclude=("b", "c")
        )
    assert str(e.value).startswith("'columns_to_include' and 'columns_to_exclude' must be a list")


def test_slice_dataframe_invalid_parameters2(request):
    dataset = request.getfixturevalue("flat_dataset")

    with pytest.raises(Exception) as e:
        slice_dataframe(input_dataframe=dataset, columns_to_include=[1, "a", None])
    assert str(e.value).startswith(
        "Member of 'columns_to_include' and" " 'columns_to_exclude' must be a string"
    )


def test_slice_dataframe_invalid_parameters3(request):
    dataset = request.getfixturevalue("flat_dataset")

    with pytest.raises(Exception) as e:
        slice_dataframe(input_dataframe=dataset, columns_to_include=["a"], columns_to_exclude=["a"])
    assert str(e.value).startswith("At least one column should be listed")



import pytest
from pyspark.sql import DataFrame, Row
from spalah.dataframe import script_dataframe


def test_script_dataframe(spark, flat_dataset: DataFrame) -> None:
    """The input dataset must be scripted and then recreated from the script:
    |  a|  b|
    +---+---+
    |  1|2.0|
    |  2|3.0|
    """
    df_input = flat_dataset.select("a", "b")

    generated_script = script_dataframe(df_input)

    # dynamically run the script which will generate "outcome_dataframe"
    loc = {"spark": spark}
    exec(generated_script, globals(), loc)

    df_outcome = loc["outcome_dataframe"]

    assert df_outcome.collect() == [Row(a=1, b=2.0), Row(a=2, b=3.0)]


def test_script_dataframe_exception_limit(spark) -> None:
    """The input dataframe has more than 20, this must raise an exception"""

    df_input = spark.range(1, 22)

    with pytest.raises(ValueError) as e:
        script_dataframe(df_input)
    assert str(e.value).startswith("This method is limited to script up")



import pytest
from spalah.dataframe import flatten_schema


@pytest.mark.parametrize(
    "assert_message,input_dataset,include_datatype,expected",
    [
        (
            "Must return all columns, including nested attributes"
            "of the complex dataset as the list",
            "nested_dataset",
            False,
            [
                "column_a",
                "column_b",
                "column_c.column_c_1",
                "column_c.column_c_2.c_2_1",
                "column_c.column_c_2.c_2_2",
                "column_c.column_c_2.c_2_3",
            ],
        ),
        (
            "Must return all columns of the flat dataset as the list",
            "flat_dataset",
            False,
            ["a", "b", "c", "d", "e"],
        ),
        (
            "Must return all columns, including nested attributes of the"
            "complex dataset as the list and the column data type",
            "nested_dataset",
            True,
            [
                ("column_a", "IntegerType()"),
                ("column_b", "DecimalType(2,1)"),
                ("column_c.column_c_1", "StringType()"),
                ("column_c.column_c_2.c_2_1", "StringType()"),
                ("column_c.column_c_2.c_2_2", "StringType()"),
                ("column_c.column_c_2.c_2_3", "StringType()"),
            ],
        ),
        (
            "Must return all columns of the flat dataset as the list and the column data type",
            "flat_dataset",
            True,
            [
                ("a", "LongType()"),
                ("b", "DoubleType()"),
                ("c", "StringType()"),
                ("d", "DateType()"),
                ("e", "TimestampType()"),
            ],
        ),
    ],
    ids=[
        "nested: without data types",
        "flat: without data types",
        "nested: with data types",
        "flat: with data types",
    ],
)
def test_flatten_schema(
    assert_message,
    input_dataset,
    include_datatype,
    expected,
    request,
) -> None:

    dataset = request.getfixturevalue(input_dataset)
    actual = flatten_schema(schema=dataset.schema, include_datatype=include_datatype)

    assert actual == expected, assert_message



from setuptools import setup, find_packages
from pathlib import Path

this_directory = Path(__file__).parent
long_description = (this_directory / "README.md").read_text()

__version__ = "0.5.0"

requirements = []

test_requirements = ['pytest>=3', ]

setup(
    author="Alex Volok",
    author_email='alexandr.volok@gmail.com',
    python_requires='>=3.7',
    classifiers=[        
        'Intended Audience :: Developers',
        'License :: OSI Approved :: MIT License',
        'Natural Language :: English',
        'Programming Language :: Python :: 3',        
        'Programming Language :: Python :: 3.7',
        'Programming Language :: Python :: 3.8',
        'Programming Language :: Python :: 3.9',
        'Programming Language :: Python :: 3.10',
    ],
    description="Spalah is a set of PySpark dataframe helpers",
    long_description=long_description,
    long_description_content_type='text/markdown',
    install_requires=requirements,
    license="MIT license",    
    include_package_data=True,
    keywords='spalah',
    name='spalah',
    packages=find_packages(include=['spalah', 'spalah.*']),
    test_suite='tests',
    tests_require=test_requirements,
    url='https://github.com/avolok/spalah',
    version=__version__,
    zip_safe=False,
)



import os
from collections import namedtuple
from delta import DeltaTable
from pyspark.sql import SparkSession
from typing import Union


def check_dbfs_mounts(mounts: list, print_output: bool = False):

    dbfs_mount = namedtuple("dbfs_mount", ("result", "mount"))
    _result = list()

    for mount in mounts:
        _mount_path = f"/dbfs{mount.mountPoint}"

    if (
        "/databricks" not in _mount_path
        and os.path.isdir(_mount_path)
        and "mount.err" not in os.listdir(_mount_path)
    ):
        _result.append(dbfs_mount("ok", mount.mountPoint))
    else:
        _result.append(dbfs_mount("failed", mount.mountPoint))

    if print_output:
        for item in _result:
            print(f"{item.result:<8}  {item.mount}")

    return _result


def get_delta_properties(
    table_path: str = "",
    table_name: str = "",
    spark_session: Union[SparkSession, None] = None,
) -> Union[dict, None]:
    """Gets dataset's delta properties as a dictionary.

    Args:
        table_path (str, optional):     Path to delta table Defaults to ''.
        table_name (str, optional):     Delta table name. For instance: db1.table1. Defaults to ''.
        spark_session: (SparkSession, optional)  The current spark context.
                                                 If not defined the getActiveSession() will be used

    Raises:
        ValueError: if values for both 'table_path' and 'table_name' provided
                    provide values to one of them
        ValueError: if values for neither 'table_path' nor 'table_name' provided
                    provide values to one of them
    """

    if table_path and table_name:
        raise ValueError("Both 'table_path' and 'table_name' provided. Use one of them.")

    if not table_path and not table_name:
        raise ValueError("Neither 'table_path' nor 'table_name' provided. Use one of them.")

    if not spark_session:
        spark_session = SparkSession.getActiveSession()

    if table_path:
        table_name = f"delta.`{table_path}`"
        _identifier_to_check = table_path
    else:
        _identifier_to_check = table_name

    if not DeltaTable.isDeltaTable(sparkSession=spark_session, identifier=_identifier_to_check):
        print(f"{table_name} is not a Delta Table")
        return None

    existing_properties = (
        spark_session.sql(f"DESCRIBE DETAIL {table_name}")
        .select("properties")
        .collect()[0]
        .asDict()["properties"]
    )

    return existing_properties


def set_delta_properties(
    properties: dict,
    table_path: str = "",
    table_name: str = "",
    keep_existing: bool = True,
    spark_session: Union[SparkSession, None] = None,
) -> None:
    """Sets and unsets delta properties. If the property already
    set with a requested value ALTER TABLE will not be triggered again

    Args:
        properties (dict):              A dictionary with properties to set.
                                        Example: {"delta.logRetentionDuration": "interval 10 days"}
        table_path (str, optional):     Path to delta table Defaults to ''.
        table_name (str, optional):     Delta table name. For instance: db1.table1. Defaults to ''.
        allow_unset (bool, optional):   If enabled, not listed properties in an arg. 'properties'
                                        but set on table will be unset. Defaults to False.
        spark_session: (SparkSession, optional)  The current spark context.
                                                 If not defined the getActiveSession() will be used

    Raises:
        ValueError: if values for both 'table_path' and 'table_name' provided
                    provide values to one of them
        ValueError: if values for neither 'table_path' nor 'table_name' provided
                    provide values to one of them
    """

    if table_path and table_name:
        raise ValueError("Both 'table_path' and 'table_name' provided. Use one of them.")

    if not table_path and not table_name:
        raise ValueError("Neither 'table_path' nor 'table_name' provided. Use one of them.")

    original_table_name = table_name

    if not spark_session:
        spark_session = SparkSession.getActiveSession()

    if table_path:
        table_name = f"delta.`{table_path}`"

    _existing_properties = get_delta_properties(
        table_name=original_table_name, table_path=table_path
    )

    if _existing_properties is not None:
        print(f"Applying table properties on '{table_name}':")

        for k, v in properties.items():

            print(f" - Checking if '{k} = {v}' is set on {table_name}")

            if k in _existing_properties and _existing_properties[k] == str(v):
                print("   Result: The property already exists on the table")
            else:
                _sql = f"ALTER TABLE {table_name} SET TBLPROPERTIES ({k} = '{v}')"
                spark_session.sql(_sql)
                print("   Result: The property has been set")

        if not keep_existing:
            for k, v in _existing_properties.items():
                if k not in properties:
                    _sql = f"ALTER TABLE {table_name} UNSET TBLPROPERTIES ({k})"
                    spark_session.sql(_sql)
                    print(
                        f"   The property '{k} = {v}' has been unset because it is not defined in "
                        "the original dict"
                    )



import pytest
from spalah.datalake import get_delta_properties


def test_get_delta_properties(spark, tmp_path) -> None:
    """Tested method must get correctly table properties as a dictionary"""

    delta_path = f"{tmp_path}/sample_delta_dataset"
    spark.range(0, 1).write.format("delta").mode("overwrite").save(delta_path)

    _sql = f"""
        ALTER TABLE delta.`{delta_path}`
        SET TBLPROPERTIES (delta.deletedFileRetentionDuration = 'interval 15 days')
    """
    spark.sql(_sql)

    existing_properties = get_delta_properties(table_path=delta_path)

    assert existing_properties == {"delta.deletedFileRetentionDuration": "interval 15 days"}


def test_set_table_properties_exceptions_both_provided_as_identifier() -> None:
    "When both table_path and table_name provided the exception must occur"

    with pytest.raises(ValueError) as e:
        get_delta_properties(table_path="/some/path", table_name="tbl_name")
    assert str(e.value).startswith("Both 'table_path' and 'table_name' provided")


def test_set_table_properties_exceptions_noting_provided_as_identifier() -> None:
    "When neither table_path nor table_name provided the exception must occur"

    with pytest.raises(ValueError) as e:
        get_delta_properties(table_path="", table_name="")
    assert str(e.value).startswith("Neither 'table_path' nor 'table_name' provided")
